;; The sequence of triangle numbers is generated by adding the natural
;; numbers. So the 7th triangle number would be 1 + 2 + 3 + 4 + 5 + 6
;; + 7 = 28. The first ten terms would be:

;; 1, 3, 6, 10, 15, 21, 28, 36, 45, 55, ...

;; Let us list the factors of the first seven triangle numbers:

;;      1: 1
;;      3: 1,3
;;      6: 1,2,3,6
;;     10: 1,2,5,10
;;     15: 1,3,5,15
;;     21: 1,3,7,21
;;     28: 1,2,4,7,14,28

;; We can see that the 7th triangle number, 28, is the first triangle
;; number to have over five divisors.

;; Which is the first triangle number to have over five-hundred
;; divisors?

(defun make-triangle-spigot ()
  (let ((sum 0)
	(num 0))
    (lambda ()
      (incf num)
      (setf sum (+ sum num)))))

;;(setf sp (make-triangle-spigot))
;;(funcall sp)

(defvar *some-primes*
  '(2 3 5 7 11 13 17 19))

(defvar *not-primes*
  '(0 1))

(defvar *max-some-primes*
  (apply #'max *some-primes*))

;; should we even assert?
(defun add-cached-prime (prime)
  (when (not (member prime *some-primes*))
    (push prime *some-primes*)
    (setf *max-some-primes* (apply #'max *some-primes*))))

(defun reset-cache ()
  (setf *max-some-primes* 2)
  (setf *some-primes* '(2 3 5)))

(defun range (start end)
  (loop for nn from start to end
	collect nn))

(defun prime? (num)
  (when (member num *not-primes*)
    (return-from prime? nil))
  (when (member num *some-primes*)
    (return-from prime? t))
  (loop for prm in *some-primes*
	do
	(when (= 0 (mod num prm))
	  (return-from prime? nil)))
  (loop for fact = (ceiling (sqrt num))
	then (if (oddp fact)
		 (- fact 2)
		 (- fact 1))
	while (>= fact *max-some-primes*)
	do
	(when (= 0 (mod num fact))
	  (return-from prime? nil)))
  t)

;; lets do some caching...
;; (reset-cache)
;; (mapcar #'add-cached-prime (remove-if-not #'prime? (range 1 10000)))
;; (setf *some-primes* (sort *some-primes* #'<))
;; *max-some-primes*

(defmacro aprog1 (first &body body)
  `(let ((it ,first))
    (progn
      ,@body)
    it))

; (macroexpand-1 '(aprog1 1 2))

(defun make-prime-spigot ()
  (let ((cached-primes *some-primes*)
	(last-num nil))
    (sort cached-primes #'<)
    (lambda ()
      (cond ((not (null cached-primes))
	     (aprog1
		 (pop cached-primes)
	       (setf last-num it)))
	    (t
	     (loop for next = (+ last-num 2)
		   then (+ next 2)
		   do
		   (when (prime? next)
		     (setf last-num next)
		     (return next))))))))


(defun find-factors (num)
  (let ((factors (list 1 num))
	(root (ceiling (/ num 2))))
    (loop for fact = 2 then (+ fact 1)
	  while (<= fact root)
	  do
	  (when (= 0 (mod num fact))
	    (push fact factors)))
    (sort factors #'<)))

;;(find-factors 100)

;; (loop for triangle-spigot = (make-triangle-spigot)
;;       for num = (funcall triangle-spigot) then (funcall triangle-spigot)
;;       for facts = (find-factors num)      then (find-factors num)
;;       while (< (length facts) 5)
;;       do
;;       (format t "~a:~a ~a~&"
;; 	      num
;; 	      (length facts)
;; 	      facts))

;(setf sp (make-triangle-spigot))
;(funcall sp)

(let ((sp (make-triangle-spigot))
      (target-len 20))
  (loop for num = (funcall sp) then (funcall sp)
	for facts = (find-factors num) then (find-factors num)
	do
	(when (> (length facts) target-len)
	  (format t "~a: ~a~&" (length facts) facts)
	  (return num))))